# -*- coding: utf-8 -*-
"""
Addon module generating constrained full-factorial DOE on 2-spaces (pi/x) problems
"""

# -------[Import necessary packages]--------------------------------------------
import os
import sys
import pyDOE2
import numpy
import math
import logging
from typing import Tuple
from inspect import isfunction
import functools
import pandas
import matplotlib.pyplot as plot
import warnings
from numpy import ndarray
from pyvplm.core.definition import PositiveParameter, PositiveParameterSet, greek_list

# -------[Global variables and settings]----------------------------------------
path = os.path.abspath(__file__)
temp_path = path.replace("\\addon\\" + os.path.basename(path), "") + "\\_temp\\"
module_logger = logging.getLogger(__name__)


# -------[Logg Exception]-------------------------------------------------------
def logg_exception(ex: Exception):
    template = "An exception of type {0} occurred. Arguments:\n{1!r}"
    message = template.format(type(ex).__name__, ex.args)
    module_logger.info(message)


# -------[Define function creating full-fact using bounds and levels]-----------
def create_doe(
    bounds: ndarray, parameters_level: ndarray, log_space: bool = True
) -> Tuple[ndarray, ndarray]:
    # noinspection PyUnresolvedReferences,PyShadowingNames
    """Functions that generates a fullfact DOE mesh using bounds and levels number.

    Parameters
    ----------
    bounds: [n*2] numpy.ndarray of floats, defines the n parameters [lower, upper] bounds

    parameters_level: [1*n] numpy.ndarray of int, defines the parameters levels

    log_space: Defines if fullfact has to be in log space or when false, linear (default is True)

    Returns
    -------
    doe_values: [m*n] numpy.ndarray of float
             A fullfact DOE, with n the number of parameters and m the number of experiments (linked to level
             repartition)

    spacing: [1*n] numpy.ndarray
          Represents the DOE's points spacing on each paramater axis in the space

    Example
    -------
    define bounds and parameters' levels:
     >>> In [1]: bounds = numpy.array([[10, 100], [100, 1000]], float)
     >>> In [2]: parameters_level = numpy.array([2, 3], int)

    generate doe in log space:
     >>> In [3]: doe_values, spacing = create_doe(bounds, parameters_level, True)

    returns:
     >>> In [4]: doe_values.tolist()
     >>> Out[4]: [[10, 100], [100, 100], [10, 316.228], [100, 316.228], [10, 1000], [100, 1000]]
     >>> In [5]: spacing.tolist()
     >>> Out[5]: [1.0, 0.5]

    """
    if (
        isinstance(bounds, numpy.ndarray)
        and isinstance(parameters_level, numpy.ndarray)
        and isinstance(log_space, bool)
    ):
        if log_space and numpy.amin(bounds) < 0:
            raise ValueError(
                "to translate on log space all bounds should be >0, else choose log_space = False."
            )
        if numpy.issubdtype(bounds.dtype, numpy.float64) and numpy.issubdtype(
            parameters_level.dtype, numpy.integer
        ):
            # Check that parameters levels syntax is correct
            if (numpy.size(parameters_level)) != (numpy.shape(bounds)[0]):
                raise ValueError("parameters_level and bounds dimensions mismatch.")
            # Check that parameters levels>=2
            if (sum(parameters_level >= 2) + sum(parameters_level == 0)) != numpy.size(
                parameters_level
            ):
                raise ValueError("parameters_level should be >=2.")
            # If log space transpose bounds in log space
            if log_space:
                bounds = numpy.log10(bounds)
            # Generate DOE on levels
            parameters_level = parameters_level + 1 * (parameters_level == 0)
            doe_levels = pyDOE2.fullfact(parameters_level).astype(int)
            for idx in range(numpy.shape(doe_levels)[1]):
                if sum(doe_levels[:, idx]) == 0:
                    doe_levels[:, idx] = 1
            # Translate levels into values x=xmin+level/max(level)*(xmax-xmin)
            doe_values = bounds[:, 0] + doe_levels / doe_levels.max(axis=0) * (
                bounds[:, 1] - bounds[:, 0]
            )
            # Calculate spacing in full-fact space (linear or log)
            spacing = 1 / doe_levels.max(axis=0) * (bounds[:, 1] - bounds[:, 0])
            # Transform calculated value from log to linear if necessary
            doe_values = 10**doe_values if log_space else doe_values
            return doe_values, spacing
        elif not (numpy.issubdtype(bounds.dtype, numpy.float64)):
            raise TypeError("elements type in in bounds should be float.")
        else:
            raise TypeError("elements type in in parameters_level should be integer.")
    elif not (isinstance(bounds, numpy.ndarray)):
        raise TypeError("bounds shoold be numpy array.")
    elif not (isinstance(parameters_level, numpy.ndarray)):
        raise TypeError("parameters_level shoold be numpy array.")
    else:
        raise TypeError("log_space shoold be boolean.")


# -------[Define function keeping nominal point if surrounded by feasible]------
def surroundings(
    doe, nominal_doe: ndarray, proper_spacing: ndarray, LogLin: bool = True
) -> Tuple[ndarray, ndarray]:
    # noinspection PyUnresolvedReferences,PyShadowingNames
    """Function to reduce a given nominal DOE on a max distance criteria with points from feasible DOE ('reachable'
    points).

    Parameters
    ----------
    doe: [m*n] numpy.ndarray of int or float
      DOE representing m feasible experiments expressed with n parameters with non-optimal spacing

    nominal_doe: [k*n] numpy.ndarray of int or float
              Fullfact DOE with k wished experiment (k<<m) expressed with the same n parameters

    proper_spacing: [n*1] numpy.ndarray of float
                 Represents max distance criteria on each DOE axis (i.e. parameter scale)

    LogLin: defines if fullfact has to be in log space or when false, linear (default is True)

    Returns
    -------
    reduced_nominal_doe: [l*n] numpy.ndarray
                      A reduced set of nominal_doe (l<=k) validating proper_spacing criteria with feasible
                      points from DOE

    to_be_removed: numpy.ndarray of bool
              Returns the corresponding indices that do not validate proper_spacing criteria

    Example
    -------
    define bounds and parameters' levels:
     >>> In [1]: bounds = numpy.array([[10, 100], [100, 1000]], float)
     >>> In [2]: parameters_level_nominal = numpy.array([2, 3], int)
     >>> In [3]: parameters_level_feasible = numpy.array([4, 6], int)

    generate doe(s) in log space:
     >>> In [4]: doe, _ = create_doe(bounds, parameters_level_feasible, True)
     >>> In [5]: nominal_doe, proper_spacing = create_doe(bounds, parameters_level_nominal, True)

    search surrounding points:
     >>> In [6]: reduced_nominal_doe, to_be_removed = surroundings(doe, nominal_doe, proper_spacing, True)
     >>> In [7]: reduced_nominal_doe.tolist()
     >>> Out[7]: [[10.0, 100.0], [100.0, 100.0], [10.0, 316.22776601683796], [100.0, 316.22776601683796],
     >>>    ...: [10.0, 1000.0], [100.0, 1000.0]]
     >>> In [8]: to_be_removed.tolist()
     >>> Out[8]: [False, False, False, False, False, False]

    """
    if (
        isinstance(doe, numpy.ndarray)
        and isinstance(nominal_doe, numpy.ndarray)
        and isinstance(proper_spacing, numpy.ndarray)
        and isinstance(LogLin, bool)
    ):
        # Proceed on type verifications
        if not (
            numpy.issubdtype(doe.dtype, numpy.float64) or numpy.issubdtype(doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in doe should be float or integer.")
        if not (
            numpy.issubdtype(nominal_doe.dtype, numpy.float64)
            or numpy.issubdtype(nominal_doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in nominal_doe should be float or integer.")
        if not (numpy.issubdtype(proper_spacing.dtype, numpy.float64)):
            raise TypeError("elements type in nominal_doe should be float or integer.")
        if not (
            numpy.shape(doe)[1] == numpy.shape(nominal_doe)[1]
            and numpy.shape(doe)[1] == len(proper_spacing)
        ):
            raise IndexError("column numbers mismatch between doe, nominal_doe and dmax.")
        # Transform DOE into log space if needed
        X = numpy.log10(doe) if LogLin else doe
        Y = numpy.log10(nominal_doe) if LogLin else nominal_doe
        # For each point in nominal_doe find if a point in doe at dmax distance in each dimension and canceled removal
        to_be_removed = numpy.ones(len(nominal_doe), bool)
        for y_idx in range(numpy.shape(Y)[0]):
            valid_distance = bool(
                numpy.sum(
                    (
                        numpy.sum((abs(X - Y[y_idx, :]) <= proper_spacing).astype(int), axis=1)
                        == len(proper_spacing)
                    ).astype(int)
                )
                >= 1
            )
            if valid_distance:
                to_be_removed[y_idx] = False
        return nominal_doe[to_be_removed == False], to_be_removed
    elif not (isinstance(doe, numpy.ndarray)):
        raise TypeError("doe should be numpy array.")
    elif not (isinstance(nominal_doe, numpy.ndarray)):
        raise TypeError("nominal_doe shoold be numpy array.")
    elif not (isinstance(proper_spacing, numpy.ndarray)):
        raise TypeError("proper_spacing shoold be numpy array.")
    else:
        raise TypeError("log_space shoold be boolean.")


# -------[Define function finding the choice_nb nearest points to nominal]------
def find_nearest(doe, nominal_doe, choice_nb, proper_spacing, log_space=True):
    # noinspection PyUnresolvedReferences
    """Function that returns for each point in nominal DOE point, the indices and max relative error for choice_nb
    nearest points in feasible DOE. As a distance has to be computed to select nearest in further functions, it is
    the max value of the relative errors (compared to the bounds) that is returned (this avoids infinite relative error
    for [0, 0] origin point).

    Parameters
    ----------
    doe: [m*n] numpy.ndarray of int or float
      DOE representing m feasible experiments expressed with n parameters with non-optimal spacing

    nominal_doe: [k*n] numpy.ndarray of int or float
              Fullfact DOE with k wished experiment (k<<m) expressed with the same n parameters

    choice_nb: int
            Number of the nearest points returned from DOE for each nominal DOE point, criteria is max relative
            distance error max(x-x_n/(max(x_n)-min(x_n)))

    proper_spacing: [n*1] numpy.ndarray of float
                     Represents max distance criteria on each DOE axis (i.e. parameter scale)

    log_space: bool
            Defines if fullfact has to be in log space or when false, linear (default is True)

    Returns
    -------
    nearest_index_in_doe: [k*choice_nb] numpy.ndarray of int
                        Gathers the corresponding 'choice_nb' nearest DOE points indices

    Example
    -------
    to define DOEs, see :func:`~sizinglab.addon.pixdoe.surroundings`

    then extract the 2 nearest feasible points for each nominal point:
     >>> In [6]: index, max_rel_distance = find_nearest(doe, nominal_doe, 2, proper_spacing, True)
     >>> In [7]: index.tolist()
     >>> Out[7]: [[0, 4], [3, 7], [8, 12], [11, 15], [20, 16], [23, 19]]
     >>> In [8]: max_rel_distance.tolist()
     >>> Out[8]: [[0.0, 0.20000000000000018], [0.0, 0.20000000000000018],
     [0.10000000000000009, 0.10000000000000009], [0.10000000000000009, 0.10000000000000009]
     [0.0, 0.20000000000000018], [0.0, 0.20000000000000018]]

    """
    if (
        isinstance(doe, numpy.ndarray)
        and isinstance(nominal_doe, numpy.ndarray)
        and isinstance(proper_spacing, numpy.ndarray)
        and isinstance(choice_nb, int)
        and isinstance(log_space, bool)
    ):
        # Proceed on type verifications
        if not (
            numpy.issubdtype(doe.dtype, numpy.float64) or numpy.issubdtype(doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in doe should be float or integer.")
        if not (
            numpy.issubdtype(nominal_doe.dtype, numpy.float64)
            or numpy.issubdtype(nominal_doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in nominal_doe should be float or integer.")
        if not (numpy.shape(doe)[1] == numpy.shape(nominal_doe)[1]):
            raise IndexError("column numbers mismatch between doe and nominal_doe.")
        if choice_nb < 1:
            raise ValueError("choice_nb numbers should be >= 1")
        # Initialise distance and index matrices
        i = 0
        nearest_index_in_doe = -1 * numpy.ones([numpy.shape(nominal_doe)[0], choice_nb], dtype=int)
        # If necessary convert data
        X = numpy.log10(doe) if log_space else doe
        Y = numpy.log10(nominal_doe) if log_space else nominal_doe
        # Find for each nominal PI value in DOE n<=choice_nb nearest points
        for x_value in Y:
            # Filter data to limit to the ones in the proper_spacing space envelope
            index = numpy.array(range(numpy.shape(X)[0]))
            reduced_index = index[
                numpy.sum((abs(X - x_value) <= proper_spacing).astype(int), axis=1)
                == len(proper_spacing)
            ]
            if len(reduced_index) <= choice_nb:
                nearest_index_in_doe[i, : len(reduced_index)] = reduced_index
            else:
                # If more than choice_nb point available, select the one with smaller relative distance
                reduced_X = X[reduced_index, :]
                rel_distance_matrix = (reduced_X - x_value) / (
                    numpy.amax(Y, axis=0) - numpy.amin(Y, axis=0)
                )
                rel_distance_vector = numpy.sum(rel_distance_matrix**2, axis=1) ** 0.5
                nearest_index_in_doe[i] = reduced_index[
                    numpy.argpartition(rel_distance_vector, choice_nb)[:choice_nb]
                ]
            i += 1
        return nearest_index_in_doe
    elif not (isinstance(doe, numpy.ndarray)):
        raise TypeError("doe shoold be numpy array.")
    elif not (isinstance(nominal_doe, numpy.ndarray)):
        raise TypeError("nominal_doe shoold be numpy array.")
    elif not (isinstance(choice_nb, numpy.ndarray)):
        raise TypeError("choice_nb shoold be an integer.")
    elif not (isinstance(proper_spacing, numpy.ndarray)):
        raise TypeError("proper_spacing shoold be numpy array.")
    else:
        raise TypeError("log_space shoold be boolean.")


# -------[Define function electing point by increasing occurrence]---------------
def elect_nearest(doe: ndarray, nominal_doe: ndarray, index: ndarray):
    # noinspection PyUnresolvedReferences,PyShadowingNames
    """Function that tries to assign for each point in nominal DOE, one point in feasible DOE elected from its
    'choice_nb' found indices. The assignments are done point-to-point electing each time the one maximizing minimum
    relative distance with current elected set. If from available indices they all are already in the set, point is
    deleted and thus: j<=k (not likely to happen).

    Parameters
    ----------
    doe: [m*n] numpy.ndarray of int or float
         DOE representing m feasible experiments expressed with n parameters with non-optimal spacing

    nominal_doe: [k*n] numpy.ndarray of int or float
                 Fullfact DOE with k wished experiment (k<<m) expressed with the same n parameters

    index: [k*nb_choice] numpy.ndarray of int
            Gathers the corresponding 'choice_nb' nearest DOE points indices (computed with :~pixdoe.find_nearest)

    Returns
    -------
    doe_elected: [j*n] numpy.ndarray of int or float
                 Returned DOE with feasible points assigned to reduced nominal DOE (deleted points with no
                 assignment, i.e. all indices already assigned)

    reduced_nominal_doe: [j*n] numpy.ndarray of int or float
                           Reduced nominal DOE (j<=k), all point are covered with feasible point

    Example
    -------
    to define DOEs and find the nearest points, see :func:`~sizinglab.addon.pixdoe.surroundings`

    then elect one point for each nominal point:
        >>> In [7]: doe_elected, reduced_nominal_doe, max_error = elect_nearest(doe, nominal_doe, index)
        >>> In [8]: doe_elected.tolist()
        >>> Out[8]: [[100.0, 100.0], [10.0, 251.18864315095797], [100.0, 251.18864315095797], [10.0, 1000.0],
        [100.0, 1000.0]]
        >>> In [9]: reduced_nominal_doe.tolist()
        >>> Out[9]: [[100.0, 100.0], [10.0, 316.22776601683796], [100.0, 316.22776601683796], [10.0, 1000.0],
        [100.0, 1000.0]]
        >>> In [10]: max_error.tolist()
        >>> Out[10]: [0.0, 0.10000000000000009, 0.10000000000000009, 0.0, 0.0]

    """
    if (
        isinstance(doe, numpy.ndarray)
        and isinstance(nominal_doe, numpy.ndarray)
        and isinstance(index, numpy.ndarray)
    ):
        # Proceed on type verifications
        if not (
            numpy.issubdtype(doe.dtype, numpy.float64) or numpy.issubdtype(doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in doe should be float or integer.")
        if not (
            numpy.issubdtype(nominal_doe.dtype, numpy.float64)
            or numpy.issubdtype(nominal_doe.dtype, numpy.integer)
        ):
            raise TypeError("elements type in nominal_doe should be float or integer.")
        if not (numpy.issubdtype(index.dtype, numpy.integer)):
            raise TypeError("elements type in index should be integer.")
        if numpy.shape(doe)[0] < numpy.amax(index):
            raise ValueError("maximum stored index is greater than doe size.")
        if numpy.shape(nominal_doe)[0] != numpy.shape(index)[0]:
            raise ValueError("nominal_doe and index should have same number of rows.")
        # Order matching by increasing number of available points
        available_index = numpy.sum(index != -1, axis=1)
        index = index[numpy.argsort(available_index), :]
        nominal_doe = nominal_doe[numpy.argsort(available_index), :]
        # Match 1-by-1 points maximizing minimum relative distance with elected_set
        index_elected = numpy.array([]).astype(int)
        doe_elected = numpy.array([]).astype(float)
        reduced_nominal_doe = numpy.array([]).astype(float)
        for nr in range(numpy.shape(index)[0]):
            # For 1st point elect first available index
            if len(doe_elected) == 0:
                index_elected = numpy.append(index_elected, index[nr, 0])
                doe_elected = doe[index[nr, 0], :]
                reduced_nominal_doe = nominal_doe[nr, :]
            # For other points elect from available index the one with maximum distance
            else:
                # Extract available index
                available_index = index[nr, :]
                available_index = available_index[available_index != -1]
                # Remove already elected index
                to_be_kept = numpy.ones(len(available_index)).astype(bool)
                for i in range(len(available_index)):
                    if available_index[i] in index_elected:
                        to_be_kept[i] = False
                available_index = available_index[to_be_kept]
                # If 1 point remaining take it, if more than one, elect it on max_min criteria
                if len(available_index) >= 1:
                    if len(available_index) > 1:
                        rel_distance = numpy.zeros(len(available_index)).astype(float)
                        doe_range = numpy.amax(doe, axis=0) - numpy.amin(doe, axis=0)
                        doe_range = doe_range + 1 * (doe_range == 0)
                        for i in range(len(available_index)):
                            try:
                                rel_distance[i] = numpy.amin(
                                    (
                                        numpy.sum(
                                            ((doe_elected - doe[available_index[i], :]) / doe_range)
                                            ** 2,
                                            axis=1,
                                        )
                                    )
                                    ** 0.5
                                )
                            except Exception as ex:
                                logg_exception(ex)
                                rel_distance[i] = (
                                    numpy.sum(
                                        ((doe_elected - doe[available_index[i], :]) / doe_range)
                                        ** 2
                                    )
                                    ** 0.5
                                )
                        available_index = available_index[numpy.argsort(-1 * rel_distance)]
                    index_elected = numpy.append(index_elected, available_index[0])
                    doe_elected = numpy.vstack([doe_elected, doe[available_index[0], :]])
                    reduced_nominal_doe = numpy.vstack([reduced_nominal_doe, nominal_doe[nr, :]])
        return doe_elected, reduced_nominal_doe
    elif not (isinstance(doe, numpy.ndarray)):
        raise TypeError("doe shoold be numpy array.")
    elif not (isinstance(nominal_doe, numpy.ndarray)):
        raise TypeError("nominal_doe shoold be numpy array.")
    else:
        raise TypeError("index shoold be numpy array.")


# -------[Define sub-function avoid script repetition, defines constrained DOEs]
# noinspection PyShadowingNames
def declare_does(
    x_Bounds: ndarray,
    x_levels: ndarray,
    parameters_constraints,
    pi_constraints,
    func_x_to_pi,
    log_space: bool = True,
):
    """Function to generate X and Pi DOE with constraints (called as sub-function script).

    Parameters
    ----------
    x_Bounds: [n*2] numpy.ndarray of floats, defines the n parameters [lower, upper] bounds

    x_levels: [1*n] numpy.ndarray of int, defines the parameters levels

    parameters_constraints, pi_constraints: functions that define parameter and Pi constraints

    func_x_to_pi: function that translates X physical values into Pi dimensionless values
                    (space transformation matrix)

    log_space: Defines if full-factorial has to be in log space or when false, linear (default is True)

    Returns
    -------
    doeX: [m*n] numpy.ndarray of float
                A full-factorial DOE, with n the number of parameters and m the number of experiments (linked to level)

    doePI: [k*n] numpy.ndarray of float
             Represents the Pi DOE's points computed from doeX and applying both X and Pi constraints (k<=m)

    """
    doe_x, _ = create_doe(x_Bounds, x_levels, log_space)
    doe_x = doe_x[apply_constraints(doe_x, parameters_constraints) == True]
    if len(doe_x) == 0:
        doePI = []
    else:
        doePI = func_x_to_pi(doe_x.tolist())
        doePI = doePI[apply_constraints(doePI, pi_constraints) == True, :]
    return doe_x, doePI


# -------[Main function: create physical points matching nominal Pi DOE]--------
# noinspection PyShadowingNames
def create_const_doe(
    parameter_set: PositiveParameterSet,
    pi_set: PositiveParameterSet,
    func_x_to_pi,
    wished_size: int,
    **kwargs
):
    # noinspection PyUnresolvedReferences,PyTypeChecker,PyShadowingNames
    """Function to generate a constrained feasible set DOE with repartition on PI not far from nominal fullfact DOE.

    Parameters
    ----------
    parameter_set: Defines the n physical parameters for the studied problem

    pi_set: Defines the k (k<n) dimensionless parameters of the problem (WARNING: no cross-validation with
            parameter_set, uses func_x_to_pi for translation)

    func_x_to_pi: Function that translates X physical values into Pi dimensionless values (space transformation matrix)

    wished_size: Is the wished size of the final elected X-DOE that represents a constrained fullfact Pi-DOE

    **kwargs: additional argumens
                 * **level_repartition** (*numpy.array* of *int*): defines the parameters levels relative repartition,
                 default is equaly shared (same number of levels)
                 * **parameters_constraints** (*function*): returns numpy.array of bool to validate each point in X-DOE,
                 default is []
                 * **pi_constraints** (*function*): returns numpy.array of bool to validate each point in Pi-DOE,
                 default is []
                 * **choice_nb** (*int*): number of returned nearest point from DOE for each nominal DOE point,
                 default is 3
                 * **spacing_division_criteria** (*int*): (>=2) defines the subdivision admitted error in Pi nominal
                 space for feasible point, default is 5
                 * **log_space** (*bool*): defines if fullfact has to be in log space or when false, linear
                 (default is log - True)
                 * **track** (*bool*): defines if the different process steps information have to be displayed
                 (default is False)
                 * **test_mode** (*bool*): set to False to show plots (default is False)

    Returns
    -------
    doeXc: [j*n] numpy.array of float
           Represents the elected feasible constrained sets of physical parameters matching spacing criteria with
           j >= whished_size

    doePIc: [j*n] numpy.array of float
            Represents the elected feasible constrained sets of dimensionless parameters matching spacing criteria
            with j >= whished_size

    Example
    -------
    define properly the parameter, pi set and transformation function:
        >>> In [1]: from pyvplm.core.definition import PositiveParameter, PositiveParameterSet
        >>> In [2]: from pyvplm.addon.variablepowerlaw import buckingham_theorem, declare_func_x_to_pi,
        reduce_parameter_set
        >>> In [3]: u = PositiveParameter('u', [1e-9, 1e-6], 'm', 'Deflection')
        >>> In [4]: f = PositiveParameter('f', [150, 500], 'N', 'Load applied')
        >>> In [5]: l = PositiveParameter('l', [1, 3], 'm', 'Cantilever length')
        >>> In [6]: e = PositiveParameter('e', [60e9, 80e9], 'Pa', 'Young Modulus')
        >>> In [7]: d = PositiveParameter('d', [10, 60], 'mm', 'Diameter of cross-section')
        >>> In [8]: parameter_set = PositiveParameterSet(u, f, l, e, d)
        >>> In [9]: parameter_set.first('u','l')
        >>> In [10]: pi_set, _ = buckingham_theorem(parameter_set, False)
        >>> In [11]: reduced_parameter_set, reduced_pi_set = reduce_parameter_set(parameter_set, pi_set, 'l')
        >>> In [12]: func_x_to_pi = declare_func_x_to_pi(reduced_parameter_set, reduced_pi_set)

    then create a complete DOE:
        >>> In [13]: doeXc, doePIc = create_const_doe(reduced_parameter_set, reduced_pi_set, func_x_to_pi, 30,
        track=True)

        .. image:: ../source/_static/Pictures/pixdoe_create_const_doe1.png
        .. image:: ../source/_static/Pictures/pixdoe_create_const_doe2.png
    """
    # Proceed on type verifications
    if (
        isinstance(parameter_set, PositiveParameterSet)
        and isinstance(pi_set, PositiveParameterSet)
        and isfunction(func_x_to_pi)
        and isinstance(wished_size, int)
    ):
        # Set additional arguments values
        level_repartition = numpy.ones(len(list(parameter_set.dictionary.keys()))).astype(int)
        parameters_constraints = []
        pi_constraints = []
        choice_nb = 3
        spacing_division_criteria = 5
        log_space = True
        track = False
        test_mode = False
        for key, value in kwargs.items():
            if not (
                key
                in [
                    "level_repartition",
                    "parameters_constraints",
                    "pi_constraints",
                    "choice_nb",
                    "spacing_division_criteria",
                    "log_space",
                    "track",
                    "test_mode",
                ]
            ):
                raise KeyError("unknown argument " + key)
            elif key == "level_repartition":
                if isinstance(value, numpy.ndarray):
                    if len(value) == len(list(parameter_set.dictionary.keys())):
                        level_repartition = value
                        for level in level_repartition:
                            if isinstance(level, int):
                                if level < 1:
                                    raise ValueError(
                                        "each level in level_repartition should be >1."
                                    )
                            else:
                                raise ValueError(
                                    "each level in level_repartition should be an integer."
                                )
                    else:
                        raise ValueError("level_repartition mismatch parameter_set keys number.")
                else:
                    raise TypeError("level_repartition should be a numpy array.")
            elif key == "parameters_constraints":
                if isfunction(value):
                    parameters_constraints = value
                else:
                    raise TypeError("parameters_constraints should be a function.")
            elif key == "pi_constraints":
                if isfunction(value):
                    pi_constraints = value
                else:
                    raise TypeError("pi_constraints should be a function.")
            elif key == "choice_nb":
                if isinstance(value, int):
                    if value <= 0:
                        ValueError("choice_nb should be >=1.")
                    else:
                        choice_nb = value
                else:
                    raise TypeError("choice_nb should be an integer.")
            elif key == "spacing_division_criteria":
                if isinstance(value, int):
                    if value <= 1:
                        ValueError("spacing_division_criteria should be >=2.")
                    else:
                        spacing_division_criteria = value
                else:
                    raise TypeError("spacing_division_criteria should be an integer.")
            elif key == "log_space":
                if isinstance(value, bool):
                    log_space = value
                else:
                    raise ValueError("log_space should be a boolean.")
            elif key == "track":
                if isinstance(value, bool):
                    track = value
                else:
                    raise ValueError("track should be a boolean.")
            elif key == "test_mode":
                if isinstance(value, bool):
                    test_mode = value
                else:
                    raise ValueError("test_mode should be a boolean.")
        # Extract bounds on parameters set and parameters number
        x_Bounds = []
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for index in parameter_set.dictionary.keys():
                # noinspection PyProtectedMember
                x_Bounds.append(parameter_set[index]._SI_bounds)
        x_Bounds = numpy.array(x_Bounds)
        parameters_number = numpy.shape(x_Bounds)[0]
        # Extract bounds on pi set and pi number
        pi_Bounds = []
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for index in pi_set.dictionary.keys():
                # noinspection PyProtectedMember
                pi_Bounds.append(pi_set[index]._SI_bounds)
        pi_Bounds = numpy.array(pi_Bounds)
        pi_number = numpy.shape(pi_Bounds)[0]
        # Check func_x_to_pi function
        try:
            pi = func_x_to_pi(numpy.transpose(x_Bounds))
            if numpy.shape(pi)[1] != len(pi_Bounds):
                raise TypeError
        except Exception:
            raise IndexError(
                "func_x_to_pi can't be used to translate physical parameters into dimensionless ones."
            )
        # Proceed on size verifications and value type
        if numpy.shape(level_repartition)[0] != parameters_number:
            raise IndexError("level_repartition index differs from parameters in parameter_set.")
        if choice_nb <= 0:
            raise ValueError("choice_nb should be >= 1.")
        if not (numpy.issubdtype(level_repartition.dtype, numpy.integer)):
            raise TypeError("level_repartition type in index should be integer.")

        # Define factorisation for population calculation

        # noinspection PyShadowingNames
        def fact_level(x):
            for idx in range(len(x)):
                if x[idx] == 0:
                    x[idx] = 1
            y = functools.reduce(lambda x, y: x * y, x)
            return y

        # Define DOE level name
        # noinspection PyShadowingNames
        def level_name(x):
            name = ""
            for idx in range(len(x)):
                if x[idx] == 0:
                    x[idx] = 1
                name += str(x[idx]) + "x"
            name = name[0 : len(name) - 1] + "=" + str(fact_level(x))
            return name

        # Save level repartition as x_levels and adapt it for constant parameter
        x_levels = level_repartition
        i = 0
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for index in parameter_set.dictionary.keys():
                # noinspection PyProtectedMember
                if parameter_set[index]._SI_bounds[0] == parameter_set[index]._SI_bounds[1]:
                    x_levels[i] = 0
                i += 1
        min_level = min(x_levels + sys.maxsize * (x_levels == 0))
        for idx in range(len(x_levels)):
            x_levels[idx] = int(1 / min_level * x_levels[idx])
        # Adapt values
        if wished_size < 2**pi_number:
            wished_size = 2**pi_number
            warnings.warn(
                "Experiments size changed to {} to obtain 2-levels full-fractional on Pi parameters".format(
                    wished_size
                )
            )
        # Create level repartition on pi and adapt it for constant parameter
        pi_levels = numpy.ones(pi_number, dtype=int)
        i = 0
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for index in pi_set.dictionary.keys():
                # noinspection PyProtectedMember
                if pi_set[index]._SI_bounds[0] == pi_set[index]._SI_bounds[1]:
                    pi_levels[i] = 0
                i += 1
        # Set an initial set point on X 3 times greater than the wished constrained set (size will be auto. adjusted)
        init_coverage_factor = 3
        x_steps = 2
        while fact_level(x_steps * x_levels) < (init_coverage_factor * wished_size):
            x_steps += 1
        x_steps += -1
        # Force entry to loop and initiate pi_level
        obtained_size_on_pi = 0
        pi_steps = 0
        save = {}
        # Starts automatic definition of initial X non-constrained set and nominal Pi set to have sufficient constrained
        # nominal Pi set
        while obtained_size_on_pi < wished_size:
            # Init variables before entering X-DOE automatic loop
            step = 1
            obtained_size_on_x = 0
            # [PHASE1] Loop increasing x parameters' level until obtaining a constrained set
            # size >= wished_size * init_coverage_factor [CAN BE SLOW]
            if track:
                print(
                    "PHASE1: Constructing constrained X-DOE based on size >= {} criteria".format(
                        wished_size * init_coverage_factor
                    )
                )
            while obtained_size_on_x < (wished_size * init_coverage_factor):
                x_steps += 1
                doeX, doePI = declare_does(
                    x_Bounds,
                    x_steps * x_levels,
                    parameters_constraints,
                    pi_constraints,
                    func_x_to_pi,
                    log_space,
                )
                save["doePI"] = doePI
                obtained_size_on_x = numpy.shape(doeX)[0]
                if track:
                    print(
                        "Step{}: non constrained {} X-DOE factorial experiment leads to constrained [{}*{}] "
                        "X-DOE matrix".format(
                            step,
                            level_name(x_steps * x_levels),
                            obtained_size_on_x,
                            parameters_number,
                        )
                    )
                    if not (obtained_size_on_x < (wished_size * init_coverage_factor)):
                        print("Skipping to PHASE2...\n")
                step += 1
            # Calculate the equivalent init_coverage_factor for obtained parameters_level such as constrained doe
            # size >= wished_size * init_coverage_factor
            init_coverage_factor = math.trunc(fact_level(x_steps * x_levels) / wished_size)
            # Init variables before entering nominal PI-DOE automatic loop
            step = 1
            previous_size = 0
            obtained_size_on_pi = 0
            pi_steps = 2
            while fact_level(pi_steps * pi_levels) < wished_size:
                pi_steps += 1
            pi_steps += -1
            # Loop increasing nominal pi parameters' level till obtaining a constrained set
            # size >= wished_size [CAN BE SLOW]
            if track:
                print(
                    "PHASE2: Constructing unconstrained Pi-DOE validating max spacing criteria = 1/{} "
                    "with feasible points".format(spacing_division_criteria)
                )
            while obtained_size_on_pi < wished_size:
                pi_steps += 1
                # If nominal pi set becomes greater than feasible pi set increase parameter level
                # (i.e. generated constrained X-DOE)
                if obtained_size_on_x < fact_level(pi_steps * pi_levels):
                    if track:
                        print(
                            "Step{}: [ERROR] Pi set size would tend to be greater than constrained X set: restart "
                            "in PHASE1 to increase X parameters levels\n".format(step)
                        )
                    break
                else:
                    doePIn, spacing = create_doe(pi_Bounds, pi_steps * pi_levels, log_space)
                    save["doePIn"] = doePIn
                    doePIn = doePIn[apply_constraints(doePIn, pi_constraints) == True]
                    save["doePIn_c"] = doePIn
                    # noinspection PyUnboundLocalVariable
                    doePIn, _ = surroundings(
                        doePI, doePIn, spacing / spacing_division_criteria, log_space
                    )
                    save["doePIn_a"] = doePIn
                    obtained_size_on_pi = numpy.shape(doePIn)[0]
                    if previous_size > obtained_size_on_pi:
                        if track:
                            print(
                                "Step{}: [ERROR] Pi set size decreasing while increasing levels {}->{}: restart "
                                "in PHASE1 to increase X parameters levels\n".format(
                                    step, previous_size, obtained_size_on_pi
                                )
                            )
                        break
                    else:
                        if track:
                            print(
                                "Step{}: non constrained [{}] Pi-DOE factorial experiment leads to constrained "
                                "[{}*{}] Pi-DOE matrix".format(
                                    step,
                                    level_name(pi_steps * pi_levels),
                                    obtained_size_on_pi,
                                    pi_number,
                                )
                            )
                previous_size = obtained_size_on_pi
                step += 1
        # From initial nominal Pi set and constrained X set extract the nearest points
        # noinspection PyUnboundLocalVariable
        index = find_nearest(
            doePI, doePIn, choice_nb, spacing / spacing_division_criteria, log_space
        )
        index_vector = (
            numpy.reshape(index, numpy.shape(index)[0] * choice_nb) if choice_nb != 1 else index
        )
        save["doePI_n"] = doePI[index_vector, :]
        # noinspection PyUnboundLocalVariable
        doeXn, doePIn = elect_nearest(doeX, doePIn, index)
        # Delete points that do not match spacing criteria
        doePI, to_be_removed = surroundings(
            doePIn, func_x_to_pi(doeXn), spacing / spacing_division_criteria, log_space
        )
        doeXn = doeXn[to_be_removed == False]
        reduction_factor = 1 - len(doeXn) / fact_level(x_steps * x_levels)
        if not test_mode:
            print("\n")
            print(
                "Set reduction factor (from feasible to optimal) is {}%\n".format(
                    round(reduction_factor * 10000) / 100
                )
            )
        # Calculate pi-DOE from elected X-DOE set
        doeXc = doeXn
        doePIc = func_x_to_pi(doeXc.tolist())
        save["doePI_e"] = doePIc
        # Plot Pi vs Pi full-factorial graphs
        if not test_mode:
            X = numpy.log10(save["doePI"]) if log_space else save["doePI"]
            X1 = numpy.log10(save["doePI_n"]) if log_space else save["doePI_n"]
            X2 = numpy.log10(save["doePI_e"]) if log_space else save["doePI_e"]
            Y = numpy.log10(save["doePIn_c"]) if log_space else save["doePIn_c"]
            Y1 = numpy.log10(save["doePIn_a"]) if log_space else save["doePIn_a"]
            x_labels = list(pi_set.dictionary.keys())
            graph_nb = 0
            for i in range(numpy.shape(Y)[1] - 1):
                for k in range(i + 1, numpy.shape(Y)[1]):
                    graph_nb += 1
            n = math.ceil(graph_nb**0.5)
            fig, axes = plot.subplots(n, n, figsize=(6 * n, 6 * n))
            graph_idx = 0
            for i in range(numpy.shape(Y)[1] - 1):
                for k in range(i + 1, numpy.shape(Y)[1]):
                    if graph_nb == 1:
                        axes_handle = axes
                    else:
                        nr = math.floor(graph_idx / n)
                        nc = graph_idx - nr * n
                        axes_handle = axes[nr, nc]
                    axes_handle.plot(X[:, i], X[:, k], "g.", label="All (Feas.)")
                    axes_handle.plot(
                        X1[:, i], X1[:, k], "c.", label="{}-nearest (Feas.)".format(choice_nb)
                    )
                    axes_handle.plot(X2[:, i], X2[:, k], "b.", label="Elected (Feas.)")
                    axes_handle.plot(Y[:, i], Y[:, k], "k.", label="All (Obj.)")
                    axes_handle.plot(Y1[:, i], Y1[:, k], "r.", label="Active (Obj.)")
                    expression = (
                        ("$log(" + x_labels[i].replace("pi", "\\pi_{") + "})$")
                        if log_space
                        else (x_labels[i].replace("pi", "$\\pi_{") + "}$")
                    )
                    axes_handle.set_xlabel(expression)
                    expression = (
                        ("$log(" + x_labels[k].replace("pi", "\\pi_{") + "})$")
                        if log_space
                        else (x_labels[k].replace("pi", "$\\pi_{") + "}$")
                    )
                    axes_handle.set_ylabel(expression)
                    axes_handle.legend()
                    ymax = max(numpy.amax(X[:, k]), numpy.amax(Y[:, k]))
                    ymin = min(numpy.amin(X[:, k]), numpy.amin(Y[:, k]))
                    xmax = max(numpy.amax(X[:, i]), numpy.amax(Y[:, i]))
                    xmin = min(numpy.amin(X[:, i]), numpy.amin(Y[:, i]))
                    try:
                        x_lines = (pi_steps - 1) * pi_levels[k] * spacing_division_criteria + 1
                        axes_handle.set_xticks(numpy.linspace(xmin, xmax, x_lines))
                    except Exception as ex:
                        logg_exception(ex)
                    axes_handle.xaxis.set_ticklabels([])
                    try:
                        y_lines = (pi_steps - 1) * pi_levels[i] * spacing_division_criteria + 1
                        axes_handle.set_yticks(numpy.linspace(ymin, ymax, y_lines))
                    except Exception as ex:
                        logg_exception(ex)
                    axes_handle.yaxis.set_ticklabels([])
                    axes_handle.grid()
                    axes_handle.set_ylim((ymin, ymax))
                    axes_handle.set_xlim((xmin, xmax))
                    graph_idx += 1
            while graph_idx < n**2:
                nr = math.floor(graph_idx / n)
                nc = graph_idx - nr * n
                axes[nr, nc].axis("off")
                graph_idx += 1
            try:
                plot.savefig(temp_path + "create_const_doe_fig1.pdf", dpi=1200, format="pdf")
            except Exception as ex:
                logg_exception(ex)
            plot.show()
            # Plot x elected vs x constrained full-fact graphs (only for variables)
            X = numpy.log10(doeXc) if log_space else doeXc
            Y = numpy.log10(doeX) if log_space else doeX
            Y_range = numpy.amax(Y, axis=0) - numpy.amin(Y, axis=0)
            Y_range = Y_range + 1 * (Y_range == 0)
            X = (X - numpy.amin(Y, axis=0)) / Y_range
            for i in range(numpy.shape(X)[1]):
                if numpy.amax(X[:, i]) == 0 and numpy.amin(X[:, i]) == 0:
                    X[:, i] = 0.5 * (X[:, i] == 0)
                    continue
            x_labels = []
            for parameter_name in parameter_set.dictionary.keys():
                if len(parameter_name.replace("_", "")) == (len(parameter_name) - 1):
                    terms = parameter_name.split("_")
                    if terms[0] in greek_list:
                        if terms[0] == terms[0].upper():
                            terms[0] = terms[0][0] + terms[0][1:].lower()
                        parameter_name = "\\" + terms[0]
                    else:
                        if terms[0].lower() in greek_list:
                            parameter_name = "\\" + terms[0].lower()
                        else:
                            parameter_name = terms[0]
                    if terms[1] in greek_list:
                        if terms[1] == terms[1].upper():
                            terms[1] = terms[1][0] + terms[1][1:].lower()
                        parameter_name += "_{\\" + terms[1] + "}"
                    else:
                        if terms[1].lower() in greek_list:
                            parameter_name += "_{\\" + terms[1].lower() + "}"
                        else:
                            parameter_name += "_{" + terms[1] + "}"
                else:
                    parameter_name = parameter_name.replace("_", "")
                    if parameter_name in greek_list:
                        if parameter_name == parameter_name.upper:
                            parameter_name = parameter_name[0] + parameter_name[:-1].lower()
                        parameter_name = "\\" + parameter_name
                x_labels.append("$" + parameter_name + "^{*}$")
            if log_space:
                title_name = "$X^{*}=\\frac{log(X)-min(log(X))}{max(log(X))-min(log(X))}$"
            else:
                title_name = "$X^{*}=\\frac{X-min(X)}{max(X)-min(X)}$"
            X_data = pandas.DataFrame(X, columns=x_labels)
            X_data["Name"] = "Feasible point"
            plot.figure(figsize=(2 * (len(x_labels) - 1), 5))
            pandas.plotting.parallel_coordinates(X_data, "Name")
            plot.xticks(fontsize=12)
            plot.title(label=title_name)
            try:
                plot.savefig(temp_path + "create_const_doe_fig2.pdf", dpi=1200, format="pdf")
            except Exception as ex:
                logg_exception(ex)
            plot.show()
        return doeXc, doePIc, save["doePI"], save["doePI_n"], save["doePIn_c"], save["doePIn_a"]
    elif not (isinstance(parameter_set, PositiveParameterSet)):
        raise TypeError("level_repartition type should be PositiveParameterSet.")
    elif not (isinstance(pi_set, PositiveParameterSet)):
        raise TypeError("pi_set type should be PositiveParameterSet.")
    elif not (isfunction(func_x_to_pi)):
        raise TypeError("func_x_to_pi should be a function.")
    else:
        raise TypeError("wished_size should be an integer.")


# -------[Wrap constraint function to avoid definition error: unconstrained]----
def apply_constraints(X: ndarray, Constraints):
    """Function to test declared constraint and return true vector if an error occurs.

    Parameters
    ----------
    X: [m*n] numpy.ndarray of float or int
       Defines the m DOE points values over the n physical parameters

    Constraints: function that should return a [1*m] numpy.ndarray of bool, that validates the m points constraint

    Returns
    -------
    Constraints(X): [1*m] numpy.ndarray of bool
                    If dimension mismatch or constraint can't be applied returns True values (no constraint applied)

    """
    # Test if some constraints are declared
    if isfunction(Constraints):
        try:  # Remove point that does not respect X constraints
            Y = Constraints(X)
            if len(Y) == len(X):
                return Constraints(X)
            else:
                print("Error applying constraints: constraints not applied!")
        except Exception as ex:
            logg_exception(ex)
            print("Error applying constraints: constraints not applied!")
    return numpy.ones(len(X), dtype=bool)



# -*- coding: utf-8 -*-
"""
Specific module to interface variablepowerlaw module with comsol FEM software

"""

# -------[Import necessary packages]--------------------------------------------
import numpy
import warnings
import re
import pandas
import pint
import logging
from numpy import ndarray
from pyvplm.core.definition import PositiveParameterSet


# -------[Define function to save doe Dataframe]--------------------------------
def save_file(
    doe_x: ndarray, file_name: str, parameter_set: PositiveParameterSet, is_SI: bool, **kwargs
):
    """Function to save .txt file within COMSOL input format.
    Values can be either expressed with SI units or user defined units (is_SI, True by default).

    Parameters
    ----------
    doe_x: DOE of the parameter_set either in defined_units or SI units

    file_name: Name of the saved file with path (example: file_name = './subfolder/name')

    parameter_set: Defines the n physical parameters for the studied problem

    is_SI: Define if parameters values are expressed in SI units or units defined by user

    """
    if (
        isinstance(doe_x, numpy.ndarray)
        and isinstance(file_name, str)
        and isinstance(parameter_set, PositiveParameterSet)
        and isinstance(is_SI, bool)
    ):
        test_mode = False
        for key, value in kwargs.items():
            if not (key in ["test_mode"]):
                raise KeyError("unknown argument " + key)
            elif key == "test_mode":
                if isinstance(value, bool):
                    test_mode = value
                else:
                    raise TypeError("test_mode should be a boolean.")
        # Check that data array and parameter set have same size
        if numpy.shape(doe_x)[1] != len(list(parameter_set.dictionary.keys())):
            raise ValueError("data dimension mismatch parameter_set keys' number")
        # Check that values are in defined bounds with 0.01% relative error flexibility
        key_list = list(parameter_set.dictionary.keys())
        for idx in range(numpy.shape(doe_x)[1]):
            max_value = numpy.amax(doe_x[:, idx])
            min_value = numpy.amin(doe_x[:, idx])
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # noinspection PyProtectedMember
                bounds = (
                    parameter_set[key_list[idx]]._SI_bounds
                    if is_SI
                    else parameter_set[key_list[idx]].defined_bounds
                )
            if (min_value < bounds[0]) and (abs(min_value - bounds[0]) / bounds[0]) < 0.0001:
                doe_x[:, idx] = numpy.maximum(doe_x[:, idx], bounds[0])
            elif (max_value > bounds[1]) and (abs(max_value - bounds[1]) / bounds[1]) < 0.0001:
                doe_x[:, idx] = numpy.minimum(doe_x[:, idx], bounds[1])
            elif (min_value < bounds[0]) or (max_value > bounds[1]):
                warnings.warn(
                    "for parameter {} saved values are out of bounds with more than 0.01% flexibility!".format(
                        key_list[idx]
                    )
                )
        # Write each parameter value on a line
        try:
            file = open(file_name + ".txt", "w")
            if not test_mode:
                print(
                    "\nREMINDER: while importing parameters' values to COMSOL use following units"
                )
            idx = 0
            for key in parameter_set.dictionary.keys():
                expression = str(key) + " "
                for value in doe_x[:, idx]:
                    expression += str(value) + ","
                file.write(expression[0 : len(expression) - 1] + "\n")
                if not test_mode:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        # noinspection PyProtectedMember
                        print(
                            "   - " + key + ": " + parameter_set[key]._SI_units
                        ) if is_SI else print(
                            "   - " + key + ": " + parameter_set[key].defined_units
                        )
                idx += 1
            file.close()
            if not test_mode:
                print("\n " + file_name + ".txt file created with success...")
        except Exception:
            raise SyntaxError("file name inappropriate, unable to open file for writing purpose.")
    else:
        if not (isinstance(doe_x, numpy.ndarray)):
            raise TypeError("data should be numpy array")
        elif not (isinstance(file_name, str)):
            raise TypeError("file_name should be a string")
        elif not (isinstance(parameter_set, PositiveParameterSet)):
            raise TypeError("parameter_set should be a PositiveParameterSet")
        else:
            raise TypeError("is_SI should be boolean")


def import_file(file_name: str, parameter_set: PositiveParameterSet, units: str):
    """Function to import .txt file generated by COMSOL (output format).
    Values can be either expressed within SI units, user defined units or specified units in the parameter name :
    'parameter_name (units)'.

    Parameters
    ----------
    file_name: str
               Name of the saved file with path (example: file_name = './subfolder/name')

    parameter_set: PositiveParameterSet
                   Defines the n physical parameters for the studied problem

    units: str
           Define what units should be considered for parameters from set
           * 'SI': means parameter is expressed within SI units, no adaptation needed and '(units)' in column name
           is ignored.
           * 'defined': means parameter is expressed within defined units written in parameter and '(units)' in
           column name is ignored, adaptation may be performed.
           * 'from_file': means parameter is expressed with file units and adaptation may be performed.

           If units are unreadable and **from_file** option is chosen, it is considered to be expressed in SI units.
           For parameters not in the defined set, if units are unreadable, there are no units or **SI** option is
           chosen, it is considered to be SI otherwise adaptation is performed.

    """
    if (
        isinstance(file_name, str)
        and isinstance(parameter_set, PositiveParameterSet)
        and isinstance(units, str)
    ):
        # Test units value
        if not (units == "SI" or units == "defined" or units == "from_file"):
            raise ValueError("units should be 'SI' or 'defined' or 'from_file'.")
        # Read file lines
        try:
            doe_x = numpy.array([]).astype(float)
            file_name = file_name + ".txt"
            empty_file = True
            labels = []
            with open(file_name, "r") as file:
                line = file.readline()
                previous_line = line
                while len(line) != 0:
                    if line[0] != "%":
                        empty_file = False
                        # Extract parameters list
                        if len(labels) == 0:
                            s = previous_line[2 : len(previous_line)]
                            labels = eval("['" + re.sub("\s+", "','", s.strip()) + "']")
                            non_units = numpy.ones(len(labels)).astype(bool)
                            for idx in range(len(labels)):
                                parameter = str(labels[idx])
                                if parameter[0] == "(":
                                    if idx != 0:
                                        labels[idx - 1] = labels[idx - 1] + " " + labels[idx]
                                    non_units[idx] = False
                            labels = numpy.array(labels)
                            labels = labels[non_units].tolist()
                        # Extract values
                        s = line
                        values = numpy.array(eval("[" + re.sub("\s+", ",", s.strip()) + "]"))
                        if len(doe_x) == 0:
                            doe_x = values
                        else:
                            doe_x = numpy.c_[doe_x, values]
                    previous_line = line
                    line = file.readline()
            if empty_file:
                doe_x = []
            else:
                doe_x = numpy.transpose(doe_x)
                if numpy.shape(doe_x)[1] != len(labels):
                    raise SyntaxError
                # Modify units name depending on chosen option
                for idx in range(len(labels)):
                    label = labels[idx]
                    parameter = label[0 : label.find("(")] if label.find("(") != -1 else label
                    if parameter in list(parameter_set.dictionary.keys()):
                        if units == "SI":
                            labels[idx] = parameter
                        elif units == "defined":
                            labels[idx] = (
                                parameter + " [" + parameter_set[parameter].defined_units + "]"
                            )
                        elif units == "from_file":
                            if label.find("(") == -1:
                                labels[idx] = parameter
                            else:
                                file_units = label[label.find("(") + 1 : len(label) - 1]
                                ureg = pint.UnitRegistry()
                                ureg.default_system = "mks"
                                Q_ = ureg.Quantity
                                try:
                                    with warnings.catch_warnings():
                                        warnings.simplefilter("ignore")
                                        value = Q_(1, file_units).to_base_units()
                                        # noinspection PyProtectedMember
                                        if str(value.units) != parameter_set[parameter]._SI_units:
                                            # noinspection PyProtectedMember
                                            raise ValueError(
                                                "dimensions mismatch for parameter {}, {} found instead of {}.".format(
                                                    parameter,
                                                    str(value.units),
                                                    parameter_set[parameter]._SI_units,
                                                )
                                            )
                                except Exception as e:
                                    labels[idx] = parameter
                                    warnings.warn(
                                        "parameter {} units defined in file are unreadable, "
                                        "SI units are applied!".format(parameter)
                                    )
                                    logging.exception(e)
                    else:
                        if label.find("(") != -1:
                            parameter = label[0 : label.find("(")]
                            file_units = label[label.find("(") + 1 : len(label) - 1]
                            ureg = pint.UnitRegistry()
                            ureg.default_system = "mks"
                            Q_ = ureg.Quantity
                            try:
                                value = Q_(1, file_units).to_base_units()
                                if units == "SI":
                                    labels[idx] = parameter + " [" + str(value.units) + "]"
                                else:
                                    labels[idx] = parameter + " [" + file_units + "]"
                            except Exception as e:
                                labels[idx] = parameter
                                warnings.warn(
                                    "parameter {} units defined in file are unreadable, it is supposed to be "
                                    "SI units!".format(parameter)
                                )
                                logging.exception(e)
                # Adapt values with units and erase in name if in parameter set
                ureg = pint.UnitRegistry()
                ureg.default_system = "mks"
                Q_ = ureg.Quantity
                for idx in range(len(labels)):
                    label = labels[idx]
                    if label.find(" [") != -1:
                        parameter = label[0 : label.find("[")]
                        parameter = parameter.replace(" ", "")
                        file_units = label[label.find("[") + 1 : len(label) - 1]
                        file_units = file_units.replace(" ", "")
                        value = Q_(1, file_units)
                        SI_value = value.to_base_units()
                        if (value.magnitude == SI_value.magnitude) and (
                            parameter in list(parameter_set.dictionary.keys())
                        ):
                            labels[idx] = parameter
                        elif (value.magnitude != SI_value.magnitude) and (
                            parameter in list(parameter_set.dictionary.keys())
                        ):
                            labels[idx] = parameter
                            for nr in range(len(doe_x[:, idx])):
                                value = Q_(doe_x[nr, idx], file_units).to_base_units()
                                doe_x[nr, idx] = value.magnitude
                        elif value.magnitude != SI_value.magnitude:
                            labels[idx] = parameter + " [" + str(SI_value.units) + "]"
                            for nr in range(len(doe_x[:, idx])):
                                value = Q_(doe_x[nr, idx], file_units).to_base_units()
                                doe_x[nr, idx] = value.magnitude
                        else:
                            labels[idx] = parameter + " [" + file_units + "]"
                doe_x = pandas.DataFrame(doe_x, columns=labels)
                # Adapt column order considering parameter_set order (add unknown parameter at the end)
                new_labels = []
                parameters_list = list(parameter_set.dictionary.keys())
                for parameter in parameters_list:
                    if parameter in labels:
                        new_labels.extend([parameter])
                    else:
                        warnings.warn("parameter {} not found in imported file!".format(parameter))
                for label in labels:
                    if not (label in new_labels):
                        new_labels.extend([label])
                doe_x = doe_x[new_labels]
            file.close()
            return doe_x
        except SyntaxError:
            raise SyntaxError(
                "parameter number and values mismatch, check that you have no spaces in parameters' names"
            )
        except Exception:
            raise SyntaxError("file name inappropriate, unable to open file for writing purpose.")
    else:
        if not (isinstance(file_name, str)):
            raise TypeError("file_name should be a string")
        elif not (isinstance(parameter_set, PositiveParameterSet)):
            raise TypeError("parameter_set should be a PositiveParameterSet")
        else:
            raise TypeError("is_SI should be a string")



# -*- coding: utf-8 -*-
"""
Core module defining elementary class and methods for SizingLab
"""

# -------[Import necessary packages]--------------------------------------------
import sys
import pint
import warnings
import logging
import sympy
import numpy
import copy
from collections import OrderedDict
from inspect import isfunction
from IPython.display import display, Math

module_logger = logging.getLogger(__name__)


# -------[Greek list definition]------------------------------------------------
# noinspection SpellCheckingInspection
greek_list = [
    "gamma",
    "delta",
    "eta",
    "theta",
    "lambda",
    "xi",
    "pi",
    "sigma",
    "upsilon",
    "phi",
    "psi",
    "omega",
]
for idx in range(len(copy.deepcopy(greek_list))):
    greek_list.append(greek_list[idx].upper())
greek_list.extend(
    [
        "alpha",
        "beta",
        "epsilon",
        "varepsilon",
        "zeta",
        "eta",
        "vartheta",
        "iota",
        "kappa",
        "mu",
        "nu",
        "rho",
        "varrho",
        "tau",
        "varphi",
        "chi",
    ]
)


# -------[Logg Exception]-------------------------------------------------------
def logg_exception(ex: Exception):
    template = "An exception of type {0} occurred. Arguments:\n{1!r}"
    message = template.format(type(ex).__name__, ex.args)
    module_logger.info(message)


# -------[Parameter Class Definition]-------------------------------------------
class Parameter:
    # noinspection PyUnresolvedReferences,GrazieInspection
    """Class defining one physical parameter.

    Attributes
    ----------
    name: str
          the parameter name as a convention will be converted to :
            - upper char(s) for constant
            - lower char(s) for variable

    defined_bounds: [1*2] list of int or float
                    converted to float and checked that defined_bounds[0]<defined_bounds[1], set to [] for constant

    value: int or float
           converted to float, set to [] for variable

    defined_units: str
                   the parameters defined units (expression checked using PINT package)

    description: str
                 some description on the parameter

    _SI_bounds (private): [1*2] list of float
                          the parameter bounds expressed into SI units automatically derived from defined bounds
                          using PINT package

    _SI_units (private): str
                         SI units are automatically derived from defined units using PINT package

    _dimensionality (private): str
                               the parameter dimensions derived from units using PINT package (ex: meter->[length])

    Examples
    --------
    save a parameter:
        >>> In [1]: m = Parameter('m', [50, 100], 'kg', 'mass')

    save a constant:
        >>> In [2]: k = Parameter('k', [2], '', 'oversizing coefficient')

    get k parameter value in equation:
        >>> In [3]: a = float(k)*2.0
        >>> Out[3]: 4.0

    print parameters'attributes:
        >>> In [4]: print(m)
        m.name=m
        m.defined_bounds=[50.0, 100.0]
        ...
        m._dimensionality=[mass]

    change parameter's attribute values:
        >>> In [5]: m.description = 'body mass'

    """

    ureg = pint.UnitRegistry()

    def __init__(self, name: str, defined_bounds: list, defined_units: str, description: str):
        """Method to create initial parameter object using syntax expressed in example."""
        # Check input types
        if not (
            isinstance(name, str)
            and isinstance(defined_bounds, list)
            and isinstance(defined_units, str)
            and isinstance(description, str)
        ):
            raise TypeError("attributes type mismatch class definition")
        # Check units syntax
        proper_syntax, formatted_units, dimensionality = self.check_units(defined_units)
        # Check bounds syntax and values
        proper_syntax = proper_syntax and self.check_bounds(defined_bounds)
        # Save parameter attributes' values
        if proper_syntax:
            if len(defined_bounds) == 2 or formatted_units == "dimensionless":
                self.name = name.lower()
            else:
                self.name = name.upper()
            if len(defined_bounds) == 2:
                lower_bound = float(defined_bounds[0])
                upper_bound = float(defined_bounds[1])
                object.__setattr__(self, "defined_bounds", [lower_bound, upper_bound])
                self.value = []
            else:
                object.__setattr__(self, "defined_bounds", [])
                self.value = defined_bounds
            object.__setattr__(self, "defined_units", formatted_units)
            self.description = description
            ureg = self.ureg
            ureg.default_system = "mks"
            Q_ = ureg.Quantity
            if len(defined_bounds) == 2:
                # noinspection PyUnboundLocalVariable
                SI_lower_bound = Q_(lower_bound, formatted_units).to_base_units()
                # noinspection PyUnboundLocalVariable
                SI_upper_bound = Q_(upper_bound, formatted_units).to_base_units()
                object.__setattr__(
                    self, "_SI_bounds", [SI_lower_bound.magnitude, SI_upper_bound.magnitude]
                )
            else:
                SI_lower_bound = Q_(defined_bounds[0], formatted_units).to_base_units()
                object.__setattr__(
                    self, "_SI_bounds", [SI_lower_bound.magnitude, SI_lower_bound.magnitude]
                )
            object.__setattr__(self, "_SI_units", str(SI_lower_bound.units))
            object.__setattr__(self, "_dimensionality", dimensionality)

    def __getattribute__(self, attribute_name):
        """Method to access parameter attribute value (for private attributes, warning is displayed).
        Access is granted using command: **parameter_name.attribute_name**.

        """
        if attribute_name in ["_SI_units", "_SI_bounds", "_dimensionality"]:
            warnings.warn("accessing private attribute value")
        return super(Parameter, self).__getattribute__(attribute_name)

    def __setattr__(self, attribute_name, value, check_units=True):
        """Method to write parameter attribute value, **parameter_name.attribute_name=value** (private attribute writing
        access denied).

        """
        if attribute_name == "defined_units":
            # Check units syntax
            proper_syntax, formatted_units, dimensionality = self.check_units(value)
            if proper_syntax:
                ureg = self.ureg
                ureg.default_system = "mks"
                Q_ = ureg.Quantity
                object.__setattr__(self, "_dimensionality", dimensionality)
                object.__setattr__(self, "defined_units", formatted_units)
                if len(self.value) == 0:
                    SI_lower_bound = Q_(self.defined_bounds[0], formatted_units).to_base_units()
                    SI_upper_bound = Q_(self.defined_bounds[1], formatted_units).to_base_units()
                else:
                    SI_lower_bound = Q_(self.value, formatted_units).to_base_units()
                    SI_upper_bound = SI_lower_bound
                object.__setattr__(
                    self, "_SI_bounds", [SI_lower_bound.magnitude, SI_upper_bound.magnitude]
                )
        elif attribute_name == "defined_bounds":
            defined_bounds = value
            # Check bounds syntax and values
            proper_syntax = self.check_bounds(defined_bounds)
            # Get units
            formatted_units = self.defined_units
            # Save parameter bounds values
            if proper_syntax:
                if len(defined_bounds) == 2:
                    object.__setattr__(self, "name", self.name.lower())
                    lower_bound = float(defined_bounds[0])
                    upper_bound = float(defined_bounds[1])
                    object.__setattr__(self, "defined_bounds", [lower_bound, upper_bound])
                    object.__setattr__(self, "value", [])
                else:
                    object.__setattr__(self, "name", self.name.upper())
                    object.__setattr__(self, "defined_bounds", [])
                    object.__setattr__(self, "value", defined_bounds)
                ureg = self.ureg
                ureg.default_system = "mks"
                Q_ = ureg.Quantity
                if len(defined_bounds) == 2:
                    # noinspection PyUnboundLocalVariable
                    SI_lower_bound = Q_(lower_bound, formatted_units).to_base_units()
                    # noinspection PyUnboundLocalVariable
                    SI_upper_bound = Q_(upper_bound, formatted_units).to_base_units()
                    object.__setattr__(
                        self, "_SI_bounds", [SI_lower_bound.magnitude, SI_upper_bound.magnitude]
                    )
                else:
                    SI_lower_bound = Q_(defined_bounds[0], formatted_units).to_base_units()
                    object.__setattr__(
                        self, "_SI_bounds", [SI_lower_bound.magnitude, SI_lower_bound.magnitude]
                    )
        elif attribute_name in [
            "name",
            "value",
            "description",
            "_SI_units",
            "_SI_bounds",
            "_dimensionality",
        ]:
            object.__setattr__(self, attribute_name, value)
        else:
            raise AttributeError("nonexistent attribute or private (write access denied)")

    def __float__(self):
        """Method to return parameter value with syntax **float(parameter_name)**.
        If value is empty (i.e. parameter is a variable), returns NaN.

        """
        if len(self.value) == 0:
            return float("nan")
        else:
            return float(self.value[0])

    def __str__(self):
        """Method used to print parameter, called with **print(parameter_name)** function."""
        statement = ""
        for key in self.__dict__.keys():
            if key[0:1] == "_":
                statement += self.name + "." + key + "(private) = " + str(self.__dict__[key]) + "\n"
            else:
                statement += self.name + "." + key + " = " + str(self.__dict__[key]) + "\n"
        return statement

    def __repr__(self):
        """Method to represent parameter definition when entering only parameter_name command."""
        return (
            self.__class__.__name__
            + " {"
            + "name:{},defined_bounds:{},value:{},defined_units:{},description:{}".format(
                self.name, self.defined_bounds, self.value, self.defined_units, self.description
            )
            + "}"
        )

    def __eq__(self, other):
        if isinstance(other, Parameter):
            return (
                self.name == other.name
                and self.defined_bounds == other.defined_bounds
                and self.value == other.value
                and self.defined_units == other.defined_units
                and self.description == other.description
            )
        return False

    def check_bounds(self, defined_bounds):
        """Method (*internal*) to check bounds syntax and value(s)."""
        proper_syntax = True
        if len(defined_bounds) == 2:
            try:
                lower_bound = float(defined_bounds[0])
                upper_bound = float(defined_bounds[1])
                if lower_bound >= upper_bound:
                    raise AssertionError
            except AssertionError:
                raise AssertionError(
                    "bad definition of bounds, should be defined_bounds[0]<defined_bounds[1]"
                )
            except Exception:
                raise TypeError("defined_bounds should be a [1x2] list of floats or integers")
        elif len(defined_bounds) == 1:
            try:
                float(defined_bounds[0])
            except Exception:
                raise TypeError("value should be a [1x1] list of float or integer")
        else:
            raise IndexError(
                "value/defined_bounds should be a [1x1] or [1x2] list of float(s)/integer(s)"
            )
        return proper_syntax

    def check_units(self, defined_units):
        """Method (*internal*) to check units value."""
        proper_syntax = True
        try:
            ureg = self.ureg
            ureg.default_system = "mks"
            Q_ = ureg.Quantity
            formatted_units = Q_("0" + defined_units)
            dimensionality = str(formatted_units.dimensionality)
            formatted_units = str(formatted_units.units)
        except Exception:
            raise ValueError(
                "bad units, type dir(pint.UnitRegistry().sys.system) with system in "
                "['US', 'cgs', 'imperial', 'mks'] for detailed list of units"
            )
        return proper_syntax, formatted_units, dimensionality


# -------[Positive Parameter Class Definition]----------------------------------
class PositiveParameter(Parameter):
    """Subclass of the class Parameter.

    Note
    ----
    This class has identical methods and parameters as the Parameter class
    except for internal check_bounds method, therefore Parameter should be
    defined with strictly positive bounds: 0<defined_bounds[0]<defined_bounds[1]

    For more details see :func:`~sizinglab.core.definition.Parameter`

    """

    def check_bounds(self, defined_bounds):
        """Method (*internal*) to check bounds syntax and value(s)."""
        proper_syntax = True
        if len(defined_bounds) == 2:
            try:
                lower_bound = float(defined_bounds[0])
                upper_bound = float(defined_bounds[1])
                if (lower_bound >= upper_bound) or (lower_bound <= 0):
                    raise AssertionError
            except AssertionError:
                raise AssertionError(
                    "bad definition of bounds, should be 0<defined_bounds[0]<defined_bounds[1]"
                )
            except Exception:
                raise TypeError("defined_bounds should be a [1x2] list of floats or integers")
        elif len(defined_bounds) == 1:
            try:
                if float(defined_bounds[0]) < 0:
                    raise AssertionError("bad definition of bounds, should be 0<defined_bounds")
            except Exception:
                raise TypeError("value should be a [1x1] list of float or integer")
        else:
            raise IndexError(
                "defined_bounds should be a [1x2] list of floats/integers such that 0<defined_bounds[0]<"
                "defined_bounds[1]"
            )
        return proper_syntax


# -------[ParameterSet Class Definition]----------------------------------------
class ParameterSet:
    # noinspection PyUnresolvedReferences
    """Class defining a set of different Parameter(s).

    Attributes
    ----------
    dictionary: OrderedDict of Parameter
                The Parameter are registered in oredered dictionary at key [Parameter.name]

    Example
    -------
    save parameters m and k:
        >>> In [1]: m = PositiveParameter('m', [50, 100], 'g', 'mass')
        >>> In [2]: K1 = Parameter('K1', [2], 'g', 'oversizing coefficient')
        >>> In [3]: parameter_set = ParameterSet(m, K1)

    add a parameter afterwards:
        >>> In [4]: K2 = Parameter('K2', [1.5], '', 'oversizing coefficient')
        >>> In [5]: parameter_set['K2'] = K2

    get K1 parameter value:
        >>> In [6]: a = float(parameter_set['K1'])

    change parameters order:
        >>> In [7]: parameter_set.first('K2', 'K1')
        >>> In [8]: print(parameter_set)
        K2: K2=1.5, oversizing coefficient
        K1: K1=2gram, oversizing coefficient
        m: m in [50.0,100.0]gram, mass

    delete K2 parameter:
        >>> In [9]: del parameter_set['K2']

    Note
    ----
    While using print function, display differs between variable and constraint.

    """

    def __init__(self, *parameters_list):
        """Method to create initial parameter set object using syntax expressed in example."""
        # Convert single parameter into tuple
        if isinstance(parameters_list, Parameter):
            parameters_list = parameters_list
        # Check input type
        if not (isinstance(parameters_list, tuple)):
            raise TypeError("parameter list should be a single Parameter or tuple")
        # Check parameters list
        proper_syntax = self.check_parameters(parameters_list)
        # Save parameters in dictionary
        if proper_syntax:
            object.__setattr__(self, "dictionary", OrderedDict())
            for i in range(len(parameters_list)):
                self.dictionary[parameters_list[i].name] = parameters_list[i]

    def __getitem__(self, index):
        """Method to return a parameter from a parameter set using its name as key:
        **parameter=parameter_set[parameter.name]**.

        """
        if index in self.dictionary.keys():
            return self.dictionary[index]

    def __setitem__(self, key, value):
        """Method to replace parameter in a parameter set or expend dictionary if new key."""
        if isinstance(key, str) and (isinstance(value, Parameter)):
            if key == value.name:
                self.dictionary[key] = value
            else:
                raise KeyError("the key mismatch parameter name")
        elif not (isinstance(key, str)):
            raise KeyError("key should be a string")
        elif not (isinstance(Parameter, str)):
            raise TypeError("assigned type should be a Parameter")

    def __delitem__(self, key):
        """Method to delete a parameter in a parameter set: **del parameter_set[parameter.name]**."""
        if key in self.dictionary.keys():
            if len(self.dictionary.keys()) == 1:
                self.dictionary = OrderedDict()
                raise Warning("empty dictionary")
            else:
                del self.dictionary[key]
        else:
            raise KeyError("the key is not in dictionary")

    def __str__(self):
        """Method used to print parameters in the set with function: **print(parameter_set)**."""
        if len(self.dictionary.keys()) == 0:
            statement = "Current set is empty"
        else:
            statement = ""
            for key in self.dictionary.keys():
                statement += key + ": " + str(self.dictionary[key].name)
                if len(self.dictionary[key].value) == 0:
                    statement += " in [" + str(self.dictionary[key].defined_bounds[0])
                    statement += "," + str(self.dictionary[key].defined_bounds[1]) + "]"
                else:
                    statement += "=" + str(self.dictionary[key].value[0])
                if not (self.dictionary[key].defined_units == "dimensionless"):
                    statement += self.dictionary[key].defined_units
                statement += ", " + self.dictionary[key].description
                statement += "\n"
        return statement

    def __eq__(self, other):
        if isinstance(other, ParameterSet):
            return self.dictionary == other.dictionary
        return False

    def latex_render(self, textArea=False):
        """Method used to print parameters in latex form: **latex_render(parameter_set)**
        When parameter name is of the form name_indice this will lead to $name_{indice}$ latex form, number is
        automatically rendered as indice.
        Greek letters will also be escaped automatically lambda_wind will lead to $\lambda_{wind}$.

        """
        logging.captureWarnings(True)
        print("Defined set is:")
        for key in self.dictionary.keys():
            key_str = str(key)
            key_str = key_str.lower()
            previous_char_is_int = False
            # Transform idx adding '_' pi1 -> pi_1
            # noinspection PyShadowingNames
            for idx in range(len(key_str)):
                char = key_str[idx]
                try:
                    int(char)
                    if not previous_char_is_int and idx != 0:
                        key_str = key_str[0:idx] + "_" + key_str[idx : len(key_str)]
                        previous_char_is_int = True
                except Exception as ex:
                    logg_exception(ex)
            # Adapt pi_1 to recognize greek expression
            if key_str.find("_") != -1 and key_str.find("_") == key_str.rfind("_"):
                key_list = key_str.split("_")
                if key_list[0] in greek_list:
                    key_list[0] = "\\" + key_list[0]
                if key_list[1] in greek_list:
                    key_list[1] = "\\" + key_list[1]
                key_str = key_list[0] + "_{" + key_list[1] + "}"
            # Adapt parameter name (if no indice used) to recognize greek expression
            if key_str in greek_list:
                key_str = "\\" + key_str
            # noinspection PyProtectedMember
            dimension = (
                ""
                if self.dictionary[key].defined_units == "dimensionless"
                else self.dictionary[key]._SI_units
            )
            if len(self.dictionary[key].defined_bounds) == 0:
                # noinspection PyProtectedMember
                expression = key_str + " = {:.2E}{}, ".format(
                    self.dictionary[key]._SI_bounds[0], dimension
                )
            else:
                # noinspection PyProtectedMember
                expression = key_str + " \in [{:.2E},{:.2E}]{}, ".format(
                    self.dictionary[key]._SI_bounds[0],
                    self.dictionary[key]._SI_bounds[1],
                    dimension,
                )
            if "*" in self.dictionary[key].description:
                try:
                    expression += (
                        key_str + "=" + sympy.latex(sympy.sympify(self.dictionary[key].description))
                    )
                except Exception as ex:
                    logg_exception(ex)
                    description = self.dictionary[key].description
                    description = description.replace(" ", "\\,")
                    expression += description
            else:
                description = self.dictionary[key].description
                description = description.replace(" ", "\\,")
                expression += description
            # added to render in TextArea
            if textArea:
                return Math(expression)
            else:
                display(Math(expression))
        logging.captureWarnings(False)
        print("")

    def __getstate__(self):
        """Method to save parameter set using pickle."""
        return self.dictionary

    # noinspection PyShadowingBuiltins
    def __setstate__(self, dict):
        """Method to read parameter set using picklerLoad()."""
        self.dictionary = dict

    def check_parameters(self, parameters_list: tuple):
        """Method (*internal*) to check parameters."""
        proper_syntax = True
        for i in range(len(parameters_list)):
            if not (isinstance(parameters_list[i], Parameter)):
                raise TypeError("all the parameters should be of Parameter type")
        return proper_syntax

    def first(self, *parameters_list):
        """Run through parameters_list tuple to move dictionary key to its position in the list."""
        # Check input type
        if not (isinstance(parameters_list, tuple)):
            raise TypeError("parameters_list should be a tuple of strings")
        if len(parameters_list) == 1 and isinstance(parameters_list[0], tuple):
            parameters_list = parameters_list[0]
        for parameter_name in parameters_list:
            if not (isinstance(parameter_name, str)):
                raise TypeError("parameters_list should be a tuple of strings")
        # Check for accessible parameters (stored)
        proper_syntax = True
        for parameter_name in parameters_list:
            if not (parameter_name in self.dictionary.keys()):
                raise KeyError("parameter {} not in parameter set".format(parameter_name))
        if proper_syntax:
            # Delete the parameters to be saved first and save it back to the end of the list
            used_keys = []
            for parameter_name in parameters_list:
                temp = self.dictionary[parameter_name]
                del self.dictionary[parameter_name]
                self.dictionary[parameter_name] = temp
                used_keys.append(parameter_name)
            # Write all the remaining parameters at the end of the dictionary
            for key in list(self.dictionary.keys()):
                if not (key in used_keys):
                    temp = self.dictionary[key]
                    del self.dictionary[key]
                    self.dictionary[key] = temp


# -------[PositiveParameterSet Class Definition]--------------------------------
class PositiveParameterSet(ParameterSet):
    """Subclass of the class Parameter.

    Note
    ----
    This class has identical methods and parameters as the Parameter class
    except for internal check_parameters method, therefore parameters_list should
    be a tuple of PositiveParameter or a single PositiveParameter

    For more details see :func:`~sizinglab.core.definition.ParameterSet`

    """

    def __setitem__(self, key, value):
        """Method to replace parameter in a parameter set or expend dictionary if new key."""
        if isinstance(key, str) and isinstance(value, PositiveParameter):
            if key == value.name:
                self.dictionary[key] = value
            else:
                raise KeyError("the key mismatch parameter name")
        elif not (isinstance(key, str)):
            raise KeyError("key should be a string")
        elif not (isinstance(PositiveParameter, str)):
            raise TypeError("assigned type should be a PositiveParameter")

    def check_parameters(self, parameters_list: tuple):
        """Method (*internal*) to check parameters."""
        proper_syntax = True
        for i in range(len(parameters_list)):
            if not (isinstance(parameters_list[i], PositiveParameter)):
                raise TypeError("all the parameters should be of PositiveParameter type")
        return proper_syntax


# -------[Constraint Class Definition]------------------------------------------
class Constraint(object):
    """Class defining a Constraint.

    Attributes
    ----------
    description: str
                 additional text to describe equation

    parameters: list(str)
                list of all the parameters names

    function: func
              is the computational expression of the constraint

    function_expr: str
                   is the literal expression of the constraint

    Example
    -------

    Note
    ----

    """

    def __init__(self, expression, desc=""):
        """Method to create initial Constraint."""
        # Forbidden character and parameters names
        forbidden_char = [
            "!",
            "$",
            "£",
            "%",
            "^",
            "#",
            "&",
            "?",
            ";",
            "ù",
            "é",
            "@",
            "¤",
            "µ",
            "è",
            "°",
            "\\",
        ]
        forbidden_param = ["I", "gamma", "beta", "re", "ln", "sqrt", "arg"]
        # Assign default values
        self.description = desc
        self.parameters = []
        self.function = None
        self.function_expr = None
        self._isAlgebraic = True
        # Adapt expression to the form expression>=0
        if not ("<=" in expression) and not (">=" in expression):
            eps = sys.float_info.min
            if "<" in expression:
                expression = expression.replace("<", "<=") + "-{}".format(eps)
            elif ">" in expression:
                expression = expression.replace(">", ">=") + "+{}".format(eps)
            else:
                raise SyntaxError("constraint expression should include inequality character.")
        if "<=" in expression:
            left = expression.split("<=")[0]
            right = expression.split("<=")[1]
            if right == "0":
                expression = "-(" + left + ")>=0"
            else:
                expression = right + "-(" + left + ")>=0"
        else:
            left = expression.split(">=")[0]
            right = expression.split(">=")[1]
            if right != "0":
                expression = left + "-(" + right + ")"
        expression = expression.strip()
        # Search for parameters
        sp_expr = sympy.sympify(expression.split(">=")[0])
        parameters = []
        if not (isinstance(sp_expr, tuple)):
            sp_expr = [sp_expr]
        for i in range(len(sp_expr)):
            symbols = sp_expr[i].free_symbols
            for symb in symbols:
                symb = str(symb)
                try:
                    if not isfunction(symb):
                        parameters.append(symb)
                    else:
                        self._isAlgebraic = False
                except Exception as ex:
                    logg_exception(ex)
                    parameters.append(symb)
        # Delete multiple data save
        if len(parameters) > 1:
            parameters = list(set(parameters))
        # Test expression
        test_expression = expression
        index = numpy.argsort(-1 * numpy.array([len(parameter) for parameter in parameters]))
        parameters = numpy.array(parameters)
        parameters = parameters[index].tolist()
        for value in parameters:
            test_expression = test_expression.replace(value, "1")
        try:
            # Check parameters names does not contain forbidden character
            for parameter in parameters:
                for char in forbidden_char:
                    if char in parameter:
                        raise ValueError("parameter names should not contain special character.")
            # Check parameters names are not forbidden names
            for parameter in parameters:
                if parameter in forbidden_param:
                    raise ValueError(
                        "parameter names should not be recognized by sympy as constant or function: {}.".format(
                            forbidden_param
                        )
                    )
            # Save parameters and expression
            self.parameters = parameters
            self.function_expr = expression
            # Define function
            expression = expression.split(">=")[0]
            s = ""
            for parameter in parameters:
                if s != "":
                    s = s + " , " + parameter
                else:
                    s = s + parameter
            self.function = eval("lambda " + s + ": (" + str(expression) + ")")
            print("step3 passed")
        except TypeError:
            raise SyntaxError("expression syntax is incorrect.")
        except Exception:
            raise SyntaxError(
                "expression error type not handled, check that none of the parameter are in forbidden set:{}.".format(
                    forbidden_param
                )
            )

    def compute(self, parameters_dict):

        parameters_values = []
        for parameter in self.parameters:
            parameters_values.append(parameters_dict[parameter])
        result = self.function(*parameters_values)
        return result

    def __str__(self):

        s = "\nConstraint : {}\n".format(self.description)
        s += "Contains {} parameter(s)\n".format(len(self.parameters))
        s += "--------------------------------------------\n"
        s += "Parameters are: ("
        for parameter in self.parameters:
            s += parameter + ","
        s = s[0 : len(s) - 1] + ")\n"
        s += "--------------------------------------------\n"
        if self._isAlgebraic:
            s += "Algebraic expression: \n \t" + self.function_expr + "\n"
        else:
            s += "Embedded function: \n \t" + self.function_expr + "\n"
        return s


# -------[ConstraintSet Class Definition]---------------------------------------
class ConstraintSet(object):
    """Class defining a ConstraintSet.

    Attributes
    ----------
    parameters: list(str)
                list of all the parameters names from all constraints

    constraints_list: list(Constraint)
                      is the list of Constraint

    Example
    -------

    Note
    ----

    """

    def __init__(self, *constraints):
        """Method to create initial ConstraintSet."""
        self.parameters = []
        self.constraints_list = []
        # save constraints and parameters
        for constraint in constraints:
            if isinstance(constraint, Constraint):
                if not (constraint in self.constraints_list):
                    self.parameters.extend(constraint.parameters)
                    self.constraints_list.append(constraint)
                else:
                    warnings.warn(
                        "trying to save the same constraint multiple times: duplicate ignored."
                    )
            else:
                warnings.warn("some of the entry are not Constraint type and will be ignored.")
        # Delete multiple data save
        if len(self.parameters) > 1:
            self.parameters = list(set(self.parameters))

    def declare_doe_constraint(self, parameter_set):
        """Specific method to generate constraint function for pyvplm pixdoe use."""
        if isinstance(parameter_set, ParameterSet):
            # check that all used parameters from constraints are declared
            for parameter in self.parameters:
                # noinspection PyUnresolvedReferences
                if not (parameter in list(ParameterSet.dictionary.keys())):
                    return []

            # write function

            def f(X):
                # check X type
                if not isinstance(X, numpy.ndarray):
                    return []
                # check X dimension
                if numpy.shape(X)[1] != len(list(parameter_set.dictionary.keys())):
                    return []
                # define function using constraints
                Y = []
                # Get pi equations
                for constraint in self.constraints_list:
                    expression = constraint.function_expr
                    # noinspection PyShadowingNames
                    idx = 0
                    # noinspection PyShadowingNames
                    for parameter in parameter_set.dictionary.keys():
                        idx += 1
                        expression = expression.replace(parameter, "X[:,{}]".format(idx))
                    Y_local = numpy.array(exec(expression)).dtype(bool)
                    if len(Y) == 0:
                        Y = Y_local
                    else:
                        Y = Y * Y_local
                return Y

            return f

    def __str__(self):

        s = "\nConstraint Set : \n"
        s += "Contains {} constraint(s)\n".format(len(self.constraints_list))
        s += "--------------------------------------------\n"
        return s



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyvplm', 'pyvplm.addon', 'pyvplm.core']

package_data = \
{'': ['*']}

install_requires = \
['ipython',
 'ipywidgets',
 'latex',
 'matplotlib',
 'numpy>=1.22.0',
 'openpyxl',
 'pandas',
 'pint',
 'pydoe2==1.3.0',
 'scikit-learn',
 'scipy',
 'sympy',
 'xlrd',
 'xlwt']

setup_kwargs = {
    'name': 'pyvplm',
    'version': '0.1.0b0',
    'description': 'Variable Power-Law regression Models tool',
    'long_description': '![](logo.png) \n\n![Tests](https://github.com/SizingLab/pyvplm/workflows/Tests/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/pyvplm/badge/?version=latest)](https://pyvplm.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n## About\n\npyVPLM is a package that is developed to help scientist, engineer, etc., to construct power-law and/or polynomial \nregression models on different type of data such as finite-element simulation results, manufacturer data-sheets...\n\nIt integrates various functionalities such as :\n\n- Model parameters reduction based on Buckingham Theorem dimensional analysis and \n- [Pint](https://pint.readthedocs.io/en/0.9/) package with derived functions.\n- Sensitivity and dependency analysis on dimensionless parameter and limited experiments to simplify further model \nexpressions.\n- Construction of optimized experimental design on feasible-physical variables leading to full-factorial design within \ndimensionless space. Those DOE are the inputs of parametrized finite-element models.\n- Regression models construction with increasing  complexity (terms sorted based on their impact) and validation based \non relative error repartition analysis.\n\n## Capabilities\n\n#### Dimensional analysis\n\nThe dimensional analysis has to be conducted on a defined set of physical parameters. It can be performed using \nalternatively `buckingham_theorem` which will return the default solution or `automatic_buckingham` which will propose \ndifferent alternate sets.\n\nBased on the obtained solutions, advanced user can also define manually a new solution set with `force_buckingham` \nfunction.\n\n```\nfrom pyvplm.core.definition import PositiveParameter, PositiveParameterSet\nfrom pyvplm.addon.variablepowerlaw import buckingham_theorem\nd = PositiveParameter(\'d\', [10e-3, 150e-3],\'m\', \'pipe internal diameter\')\ne = PositiveParameter(\'e\', [.1e-3, 10e-3], \'m\', \'pipe thickness\')\nparameter_set = PositiveParameterSet(d,e)\npi_set, _ = buckingham_theorem(parameter_set, track=False)\n```\n\n#### Sensitivity & dependency analysis\n\nOnce dimensional analysis is done, there may be still a huge number of dimensionless parameter to describe a \nperformance criteria (mostly form factor) and DOE construction phase may lead to big experiments number and long \nsimulation times.\n\nThis is to answer this problematic that `pi_sensitivity` and `pi_dependency` functions have been designed. The obtained \ngraph for analysis are based on primary vs. secondary parameters analysis that can be easily adapted using \nconfiguration parameters:\n\n![](./docs/source/_static/Pictures/variablepowerlaw_pi_sensitivity.png)\n\n------\n\n![](./docs/source/_static/Pictures/variablepowerlaw_pi_dependency.png)\n\n#### Optimized design of experiments\n\nThe non-constrained nor reduced experimental set are defined using [pyDOE2](https://github.com/clicumu/pyDOE2) package. \nIt integrates automatic sizing of physical/dimensionless initial test plans and functions for selection based on \ndistance criteria (dimensionless mapping) and spread quality (physical mapping).\n\n![](./docs/source/_static/Pictures/pixdoe_create_const_doe1.png)\n\n------\n\n![](./docs/source/_static/Pictures/pixdoe_create_const_doe2.png)\n\n#### Regression models construction\n\nThe `regression_models` function interpolate results to fit a given order polynomial model within linear or logarithmic \nspace. \n\nWithin log space, the model obtained can be transformed into variable power-law model, indeed:\n\n$$\n\\begin{align}\n    log(\\pi_0) = a_0+a_1 \\cdot log(\\pi_1) + a_{11} \\cdot log(\\pi_1)^2+a_{12} \\cdot log(\\pi_1) \\cdot log(\\pi_2) + a_2 \\cdot log(\\pi_2) +...\n\\end{align}\n$$\n\nCan be expressed in the following form:\n\n$$\n\\begin{align}\n    \\pi_0 = 10^{a_0} \\cdot \\pi_1 ^{a_1 + a_{11} \\cdot log(\\pi_1)+a_{12} \\cdot log(\\pi_2)+...} \\cdot  \\pi_2^{a_2+...} \\cdot ...\n\\end{align}\n$$\n\nThis is the origin of package name since variable power-law model is one of the current research subject of MS2M team \nin [ICA](http://institut-clement-ader.org/home/) Laboratory (Toulouse-France). \n\nRegression coefficients are sorted with increasing magnitude while considering standardized values regression (first \norder terms are selected at the beginning to avoid singularity issues):\n\n![](./docs/source/_static/Pictures/variablepowerlaw_regression_models1.png)\n\n<u>The four criteria to evaluate model fidelity with leave-one-out cross-validation are:</u>\n\n1. Maximal relative error magnitude\n2. Average value of the magnitude of relative error which is a good indicator of both average and standard deviation\n3. Average value of the relative error\n4. Standard deviation of the relative error\n\nOn this example with 2 dimensionless parameters and order 3 polynomial expression, a 5-terms model seems to have good \nrepresentation capabilities.\n\n------\n\nOnce regression models have been constructed, each one of them can be analyzed through the analysis of their relative \nerror using `perform_regression` function:\n\n![](./docs/source/_static/Pictures/variablepowerlaw_perform_regression1.png)\n\n## Examples and notes\n\nFour Jupyter Notebooks have been developed to present tool capabilities and functions. They can be launched \nusing Jupyter Notebook application and opening `.ipynb` files from `master/notebooks` folder.\n\nAdditional documentation on sub-packages (`pyvplm.core`, `pyvplm.addon`) and functions can be found on the online \n[readthedocs](https://pyvplm.readthedocs.io/en/latest/) documentation.\n\n## Install\n\nTo install pyVPLM, simply run:\n\n`pip install pyvplm`\n\n\n## Credit\n\npyVPLM is an adaptation of the work performed by MS2M team at [ICA Laboratory](http://institut-clement-ader.org/) - \nFrance and covers the work done during different doctorate thesis:\n\n- Copyright (C) 2014 - 2017 - [Florian Sanchez](https://www.linkedin.com/in/florian-sanchez-7b65ba43/)\n- Copyright (C) 2017 - 2019 - \n[Francesco De Giorgi](https://www.linkedin.com/in/francesco-de-giorgi/?originalSubdomain=fr)\n\n## Author\n\n[A. Reysset](https://www.researchgate.net/profile/Aurelien-Reysset-2)\n\n## References\n\n- F. Sanchez, M. Budinger, I. Hazyuk, "*Dimensional analysis and surrogate models for thermal modeling of power \n- electronic components*", Electrimacs conference (2017), Toulouse\n- F. Sanchez, M. Budinger, I. Hazyuk, "*Dimensional analysis and surrogate models for the thermal modeling of \nMulti-physics systems*", \n[Applied Thermal Engineering](https://www.researchgate.net/journal/1359-4311_Applied_Thermal_Engineering) \n110 (August 2016)\n\n',
    'author': 'Aurélien REYSSET',
    'author_email': 'aurelien.reysset@insa-toulouse.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SizingLab/pyvplm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
